using System;
using System.Drawing;
using System.Collections;
using System.Diagnostics;
using System.Net;
using System.IO;
using WebCamService;

// TODO:
//  Protect some public variables

namespace WebFlow
{
    public class StationInfo
    {
        public StationInfo(string strID, string strDescription)
        {
            this.strID          = strID;
            this.strDescription = strDescription;
        }

        public void SetXY(int iX, int iY)
        {
            this.iX = iX;
            this.iY = iY;
        }

        public void SetPoly(Point[] oPoints)
        {
            this.oPoints = oPoints;
        }

        // < 15%	= Green
        // 15-21%	= Yellow
        // 22-34%	= Red
        // > 35%	= Flashing
        public void SetData(int iScans, int iVolume, int iLoops)
        {
            this.iScans		= iScans;
            this.iVolume	= iVolume;		// vehicles per minute
            this.iLoops		= iLoops;

            iOccupancy = iScans / 12;

            if (iOccupancy == 0)
                colorCongestion = Color.White;
            else if (iOccupancy < 15)
                colorCongestion = Color.Green;
            else if (iOccupancy < 22)
                colorCongestion = Color.Yellow;
            else if (iOccupancy < 35)
                colorCongestion = Color.Red;
            else
                colorCongestion = Color.Black;
        }

        public string   strID;
        public string   strDescription;
        public int      iX;
        public int      iY;
        public Point[]  oPoints;
        public Color    colorCongestion;

        private int     iOccupancy;

        private int     iScans;
        private int     iVolume;
        private int     iLoops;
    }

    public class TrafficInfo
    {
        public const string strUsername             = "anonymous";
        public const string strPassword             = "user@domain.com";

        public const string strWebFlowSTA           = "WEBFLOW.STA";
        public const string strWebFlowXY            = "SysVert.xy";
        public const string strWebFlowBMP           = "SysVert.bmp";
        public const string strWebFlowDAT           = "WEBFLOW.DAT";
        public const string strWebFlowMSG           = "WEBFLOW.MSG";
        public const string strWebFlowUpdates       = "WebFlowUpdates";
        public const string strWebFlowCompressedSTA = "Wireless044.STA";
        public const string strWebFlowCompressedDAT = "Wireless044.DAT";

        public ArrayList oStations = new ArrayList();
        public int iVersion = 0;
        public int iNumStations = 0;
        public DateTime dtLastUpdated;
        public int iExpressLanes;

        private string ValidPath(string strIn)
        {
            if (!strIn.EndsWith("\\"))
                strIn += '\\';
            return strIn;
        }

        private string ValidURL(string strIn)
        {
            if (!strIn.EndsWith("/"))
                strIn += '\\';
            return strIn;
        }

        #region "Get" Commands

        private bool GetFTPFiles(string strServer, string[] strFTPFilenames, string[] strLocalFilenames)
        {
            bool fSuccess = false;

            if (strFTPFilenames.Length == strLocalFilenames.Length)
            {
                try
                {
                    FtpClient oFTPClient = new FtpClient(strServer, strUsername, strPassword, 10, 21);

                    oFTPClient.Login();
                    oFTPClient.BinaryMode = true;

                    for(int i=0; i < strFTPFilenames.Length; i++)
                    {
                        oFTPClient.Download(strFTPFilenames[i], strLocalFilenames[i]);
                    }

                    fSuccess = true;
                }
                catch(Exception x)
                {
                    Trace.WriteLine("ERROR: GetFTPFile " + x.Message);
                }
            }

            return fSuccess;
        }

        private bool GetWebFile(string strURL, string strLocalFilePath)
        {
            bool fSuccess = false;

            try
            {
                Trace.WriteLine("Making request with following URL: " + strURL);

                HttpWebRequest wReq = (HttpWebRequest)WebRequest.Create(strURL);
                wReq.Timeout = 10 * 1000;
                HttpWebResponse wResp = (HttpWebResponse)wReq.GetResponse();

                BinaryReader reader = new BinaryReader(wResp.GetResponseStream());
                byte[] strResponse = reader.ReadBytes((int)wResp.ContentLength);
                reader.Close();

                FileStream	    fs      = new FileStream(strLocalFilePath, FileMode.Create);
                BinaryWriter    writer  = new BinaryWriter(fs);
                writer.Write(strResponse);
                writer.Close();

                wResp.Close();

                fSuccess = true;
            }
            catch(WebException xWeb)
            {
                Trace.WriteLine("ERROR: GetWebFile " + xWeb.Status + " " + xWeb.Message);
            }
            catch(Exception x)
            {
                Trace.WriteLine("ERROR: GetWebFile " + x.Message);
                throw x;
            }

            return fSuccess;
        }

        public bool GetFTPTrafficView(string strServer, string strLocalBaseDir)
        {
            string[] strFTPFilenames    = new string[3] {   strWebFlowUpdates + "/" + strWebFlowBMP,
                                                            strWebFlowSTA,
                                                            strWebFlowUpdates + "/" + strWebFlowXY      };
            string[] strLocalFilenames  = new string[3] {   ValidPath(strLocalBaseDir) + strWebFlowBMP,
                                                            ValidPath(strLocalBaseDir) + strWebFlowSTA,
                                                            ValidPath(strLocalBaseDir) + strWebFlowXY   };

            return GetFTPFiles(strServer, strFTPFilenames, strLocalFilenames);
        }

        public bool GetFTPTrafficData(string strServer, string strLocalBaseDir)
        {
            string[] strFTPFilenames    = new string[2] {   strWebFlowDAT,
                                                            strWebFlowMSG                               };
            string[] strLocalFilenames  = new string[2] {   ValidPath(strLocalBaseDir) + strWebFlowDAT,
                                                            ValidPath(strLocalBaseDir) + strWebFlowMSG  };

            return GetFTPFiles(strServer, strFTPFilenames, strLocalFilenames);
        }

        public bool GetWebTrafficBMP(string strBaseURL, string strLocalBaseDir)
        {
            return GetWebFile(ValidURL(strBaseURL) + strWebFlowUpdates + "/" + strWebFlowBMP, ValidPath(strLocalBaseDir) + strWebFlowBMP);
        }

        public bool GetWebTrafficXY(string strBaseURL, string strLocalBaseDir)
        {
            return GetWebFile(ValidURL(strBaseURL) + strWebFlowUpdates + "/" + strWebFlowXY, ValidPath(strLocalBaseDir) + strWebFlowXY);
        }

        public bool GetWebTrafficSTA(string strBaseURL, string strLocalBaseDir)
        {
            return GetWebFile(ValidURL(strBaseURL) + strWebFlowSTA, ValidPath(strLocalBaseDir) + strWebFlowSTA);
        }

        public bool GetWebTrafficDAT(string strBaseURL, string strLocalBaseDir)
        {
            return GetWebFile(ValidURL(strBaseURL) + strWebFlowDAT, ValidPath(strLocalBaseDir) + strWebFlowDAT);
        }

        public bool GetWebTrafficMSG(string strBaseURL, string strLocalBaseDir)
        {
            return GetWebFile(ValidURL(strBaseURL) + strWebFlowMSG, ValidPath(strLocalBaseDir) + strWebFlowMSG);
        }

        public bool GetWebTrafficCompressedDAT(string strBaseURL, string strLocalBaseDir)
        {
            return GetWebFile(ValidURL(strBaseURL) + strWebFlowCompressedDAT, ValidPath(strLocalBaseDir) + strWebFlowCompressedDAT);
        }

        public bool GetWebTrafficCompressedSTA(string strBaseURL, string strLocalBaseDir)
        {
            return GetWebFile(ValidURL(strBaseURL) + strWebFlowCompressedSTA, ValidPath(strLocalBaseDir) + strWebFlowCompressedSTA);
        }

        #endregion "Get" Commands

        public bool ProcessStationInfo(string strBasePath)
        {
            bool fSuccess = false;

            try
            {
                oStations.Clear();
                iVersion = 0;
                iNumStations = 0;

                string strPath = ValidPath(strBasePath);

                fSuccess = ProcessSTA(strPath + strWebFlowSTA);
                fSuccess = ProcessXY(strPath + strWebFlowXY);
                fSuccess = ProcessDAT(strPath + strWebFlowDAT);

                fSuccess = true;
            }
            catch (Exception x)
            {
                Trace.WriteLine("ERROR: ProcessStationInfo " + x.Message);
            }

            return fSuccess;
        }

        public bool ProcessXY(string strPath)
        {
            bool fSuccess = false;

            try
            {
                String strInput;
                string[] strFields;
                int iCount = 0;
                StreamReader sr = File.OpenText(strPath);

                while ((strInput = sr.ReadLine()) != null) 
                {
                    if (strInput.StartsWith("ES-"))
                    {
                        strFields = strInput.Split(';');
                        string[] strInfo = strFields[0].Split(',');

                        string strThisID = strInfo[0];
                        int x = Convert.ToInt32(strInfo[1]);
                        int y = Convert.ToInt32(strInfo[2]);
                        int iNumPoints = Convert.ToInt32(strFields.Length - 1);

                        Point[] oPoints = new Point[iNumPoints];

                        for(int i = 0; i < iNumPoints; i++)
                        {
                            string[] strPoints = strFields[i+1].Split(',');
                            int xPoint = Convert.ToInt32(strPoints[0]);
                            int yPoint = Convert.ToInt32(strPoints[1]);

                            oPoints[i] = new Point(xPoint,yPoint);
                        }

                        StationInfo oStation = (StationInfo)oStations[iCount++];
                        oStation.SetXY(x, y);
                        oStation.SetPoly(oPoints);
                    }
                }
                sr.Close();

                fSuccess = true;
            }
            catch(Exception x)
            {
                Trace.WriteLine("ERROR: ProcessXY " + x.Message);
            }

            return fSuccess;
        }

        public bool ProcessSTA(string strPath)
        {
            bool fSuccess = false;

            try
            {
                String strInput;
                string[] strFields;
                StreamReader sr = File.OpenText(strPath);

                strFields = sr.ReadLine().Split('=');
                iVersion = Convert.ToInt32(strFields[1].Trim());

                strFields = sr.ReadLine().Split('=');
                iNumStations = Convert.ToInt32(strFields[1].Trim());

                while ((strInput = sr.ReadLine()) != null) 
                {
                    if (strInput.CompareTo("END") != 0)
                    {
                        strFields = strInput.Split(',');
                        string strThisID = strFields[0];
                        string strDescription = strFields[1].Trim();

                        oStations.Add(new StationInfo(strThisID, strDescription));
                    }
                }
                sr.Close();

                fSuccess = true;
            }
            catch(Exception x)
            {
                Trace.WriteLine("ERROR: ProcessSTA " + x.Message);
            }

            return fSuccess;
        }

        public bool ProcessDAT(string strPath)
        {
            bool fSuccess = false;

            try
            {
                FileStream		fs	= new FileStream(strPath, FileMode.Open, FileAccess.Read);
                BinaryReader	r	= new BinaryReader(fs);

                int iDATVersion	= r.ReadByte();	// byte 0: version number of the WEBFLOW.STA file needed to decode it
                if (iDATVersion == iVersion)
                {
                    byte express	= r.ReadByte();	// byte 1: directional data for the I90 and I5 reversible express lanes
                    //	HI(express) = I90		LO(express) = I5
                    //	0 = Closed				0 = Closed
                    //	1 = East				1 = North
                    //	2 = West				2 = South

                    byte res1		= r.ReadByte();	// byte 2: reserved
                    byte res2		= r.ReadByte();	// byte 3: reserved

                    byte year		= r.ReadByte();	// byte 4: year
                    byte month		= r.ReadByte();	// byte 5: month
                    byte day		= r.ReadByte();	// byte 6: day
                    byte hour		= r.ReadByte();	// byte 7: hour
                    byte minute		= r.ReadByte();	// byte 8: minute
                    byte second		= r.ReadByte();	// byte 9: second

                    int iCalcedCheckSum = iDATVersion + express + res1 + res2 + year + month + day + hour + minute + second;

                    DateTime oDate = new DateTime(2000 + year, month, day, hour, minute, second);

                    // Starting with byte 10, the data is in blocks of 5 bytes
                    // The number of blocks is equal to the station count (see Line 2, WEBFLOW.STA).  
                    for (int i=0; i < iNumStations; i++)
                    {
                        //
                        //	Each 5-byte block contains bit-mapped data in the following format:
                        //
                        //	0			1			2			3			4
                        //	00000000	00000000	00000000	00000000	00000000
                        //	^^^^^^^^	^^^^^^^^	^^^^^^^^	^^^^^^^^	^^^^^^^^
                        //	 |     |	|  ||  |	       |	       |	       |-	LO(Volume) (total count of vehicles)
                        //	 |     |	|  ||  |	       |	       |-------------	HI(Volume)
                        //	 |     |	|  ||  |	       |-------------------------	LO(Scan Count) (% Occupancy = Scan Count / 12)
                        //	 |     |	|  ||  |-------------------------------------	HI(Scan Count)
                        //	 |     |	|  ||----------------------------------------	Reserved
                        //	 |     |	|  |-----------------------------------------	# of loops at station (allows for several lanes per station, 0-7)
                        //	 |     |	|--------------------------------------------	Data Validity: 0=data NOT usable, 1=data OK
                        //	 |     |-------------------------------------------------	# of 20 second periods per sample (3 = 1 minute, standard)
                        //	 |-------------------------------------------------------	Incident (0=No incident, 1=Tentative, 2=Occurred, 3=Continuing)
                        //
                        byte byte0		= r.ReadByte();
                        byte byte1		= r.ReadByte();
                        byte byte2		= r.ReadByte();
                        byte byte3		= r.ReadByte();
                        byte byte4		= r.ReadByte();

                        iCalcedCheckSum += (byte0 + byte1 + byte2 + byte3 + byte4);

                        bool fValid		= Convert.ToBoolean((byte1 & 0xFF) >> 7);

                        //if (fValid)
                        {
                            byte incident	= (byte)((byte0 & 0xFF) >> 6);
                            byte periods	= (byte)((byte0 & 0x3F));
                            byte loops		= (byte)((byte1 & 0x7F) >> 4);

                            byte scanHI		= (byte)((byte1 & 0x07));
                            byte scanLO		= byte2;

                            byte volumeHI	= byte3;
                            byte volumeLO	= byte4;

                            int  iScans		= ((byte1 & 0x07) << 8) | byte2;
                            int  iVolume	= (byte3 << 8) | byte4;

                            StationInfo oStation = (StationInfo)oStations[i];
                            oStation.SetData(iScans, iVolume, loops);
                        }
                    }

                    // 4-byte arithmetic checksum, low byte is first, which is the sum of all the other bytes in this file
                    byte[] checksum = r.ReadBytes(4);
                    int iNeededChecksum = (checksum[3] << 24) | (checksum[2] << 16) | (checksum[1] << 8) | checksum[0];

                    fSuccess = (iCalcedCheckSum == iNeededChecksum);
                }

                r.Close();
                fs.Close();
            }
            catch(Exception x)
            {
                Trace.WriteLine("ERROR: ProcessDAT " + x.Message);
            }

            return fSuccess;
        }

        //
        //  BUGBUG: sysvert.xy does not map to "wirelessXXX.*"
        //  TODO: Convert sysvert.xy to wirelessXXX.xy
        //
        public bool ProcessCompressedStationInfo(string strBasePath)
        {
            bool fSuccess = false;

            try
            {
                oStations.Clear();
                iVersion = 0;
                iNumStations = 0;

                string strPath = ValidPath(strBasePath);

                fSuccess = ProcessSTA(strPath + strWebFlowCompressedSTA);

                fSuccess = ProcessXY(strPath + strWebFlowXY);

                fSuccess = ProcessCompressedDAT(strPath + strWebFlowCompressedDAT);

                fSuccess = true;
            }
            catch (Exception x)
            {
                Trace.WriteLine("ERROR: ProcessCompressedStationInfo " + x.Message);
            }

            return fSuccess;
        }

        public bool ProcessCompressedDAT(string strPath)
        {
            bool fSuccess = false;

            try
            {
                FileStream		fs	= new FileStream(strPath, FileMode.Open, FileAccess.Read);
                BinaryReader	r	= new BinaryReader(fs);

                byte[] bRawBuffer = r.ReadBytes((int)fs.Length);

                fSuccess = ExpandHuffmanData(bRawBuffer);

                r.Close();
                fs.Close();
            }
            catch(Exception x)
            {
                Trace.WriteLine("ERROR: ProcessCompressedDAT " + x.Message);
            }

            return fSuccess;
        }

        //
        // Converted to C# from:
        //  http://www.wsdot.wa.gov/PugetSoundTraffic/palmflow/huffman.c
        //
        private bool ExpandHuffmanData(byte[] bBuffer)
        {
            byte    bCurrentData;
            int     iBit = 7;
            int     iLoop = 0;
            int     iByteCount = 0;

            // Get ExpressLane Direction
            iExpressLanes = bBuffer[++iByteCount];
	
            int iYear = (bBuffer[++iByteCount] << 8) + bBuffer[++iByteCount];
            int iMonth = bBuffer[++iByteCount];
            int iDay = bBuffer[++iByteCount];
            int iHour = bBuffer[++iByteCount];
            int iMinute = bBuffer[++iByteCount];

            dtLastUpdated = new DateTime(iYear, iMonth, iDay, iHour, iMinute, 0);

            while(iLoop < iNumStations)
            {
                StationInfo oStation = (StationInfo)oStations[iLoop++];

                if (iBit < 0)
                {
                    iBit = 7;
                    iByteCount++;
                }
			
                bCurrentData = (byte)(bBuffer[iByteCount] & (0x01 << iBit));	
                iBit--;	
		
                // 0
                if (bCurrentData == 0x00)
                    //0 is signed to WIDE_OPEN on the proxy side
                    oStation.colorCongestion = Color.Green;
                    // 1
                else
                {
                    if(iBit < 0)
                    {
                        iBit = 7;
                        iByteCount++;
                    }
			
                    bCurrentData = (byte)(bBuffer[iByteCount] & (0x01 << iBit));	
                    iBit--;
			
                    //1 0
                    if (bCurrentData == 0x00)
                        // 10 is signed to NO_DATA on the proxy
                        oStation.colorCongestion = Color.Gray;
                        //1 1 
                    else 
                    {
                        if(iBit < 0)
                        {
                            iBit = 7;
                            iByteCount++;
                        }
								
                        bCurrentData = (byte)(bBuffer[iByteCount] & (0x01 << iBit));	
                        iBit--;
				
                        // 1 1 0
                        if (bCurrentData == 0x00)
                            //110 is signed to MODERATE on proxy
                            oStation.colorCongestion = Color.Yellow;
                            // 1 1 1
                        else
                        {
                            if(iBit < 0)
                            {
                                iBit = 7;
                                iByteCount++;
                            }
									
                            bCurrentData = (byte)(bBuffer[iByteCount] & (0x01 << iBit));	
                            iBit--;
					
                            // 1 1 1 0
                            if (bCurrentData == 0x00)
                                //1110 is signed to HEAVY on proxy
                                oStation.colorCongestion = Color.Red;
                                // 1 1 1 1
                            else
                                //1111 is signed to STOP_GO on proxy
                                oStation.colorCongestion = Color.Black;
                        }
                    }		
                }			
            }

            return (iLoop == iNumStations);
        }
    }
}