#!/usr/bin/perl

#---------------------------------------------------------
#GPS Web Page Sample Code by J.P. Stewart (jp@jpstewart.org)
#This Code requires a perl capable web server as well as write access to a file.
#Additionally, this script will not read from a GPS device directly.
#See the additional sample code I have previously provided to do that.

#The arguments (passed in a get query string) for this script are:
#   lat : This is the latitude component of your location in decimal
#   long: This is the longitude component of your location in decimal
#   time: This is the time in which you were at that position (just a string)

#This code can be freely used as long as this section is in tact.
#---------------------------------------------------------


#You can replace with full path to file as well. Needs write permission for the user 
# running the script. You should ensure that only you have permission to write to this
# file, or only you have permission to access this script.

$LocationStorageFileName = "location.txt";



# End of user configurable section
#_________________________________

#Minor security precaution. This should prevent most malicious attacks trying to fill up 
# your location file.
exit if (length($ENV{'QUERY_STRING'}) > 256);

#Tell the browser the format we are returning
print "Content-type: text/html\n\n";

#Check to see if there are arguments to read in (This code only supports GET arguments)
if (length($ENV{'QUERY_STRING'})>0) {

	#This code block takes each argument pair and does work on it before storing it 
	# in a hash table
	map {
		#Change pluses to spaces
		s/\+/ /g;

		#Change encoded characters like "%20" to their corresponding characters
		#Note: Because this comes first, dont send "=" encoded and not as a part 
		#      of a value or keyword.
		s/%([a-fA-F0-9][a-fA-F0-9])/pack("C",hex($1))/ge;

		#Determine the keyword and value for thie particular argument
		my ($keyword,$value) = split /=/;

		#Create a hashtable of all the arguments passed.
		$params{$keyword} = $value;

	} split(/&/,$ENV{'QUERY_STRING'});

	#If the user passed in all of the attributes we are interested in, 
	# then process them.
	if ((exists $params{lat}) && (exists $params{long}) && (exists $params{time})) {

		#Try to open the file and overwrite it. An error here will print an 
		# error to the client. Note: This is where you will hit most permission 
		# problems if you have any.
		open(FILE,">$LocationStorageFileName") or 
			&errorout("Could not write to the location file. ".
			          "Check that you have given write permissions".
				  " for the proper user. Error: $!");
		
		#Wait till no one else is using the file
		#This "could" lead to deadlock scenarios, ensure your Web Server can 
		# terminate long running scripts or look at using LOCK_NB with flock.
		flock FILE, 2;

		#Print out the data to the file
		print FILE "$params{lat}\n$params{long}\n$params{time}\n";
		
		#Close the file
		close(FILE);

		#Let everyone else waiting know we are done with the file.
		flock FILE, 8;

		#Leave the script since we are done
		exit;
	}
}

#No valid arguments were passed in, so lets print out the location.

#Setup standard HTML layout. This is basic, feel free to use your own.
print "<html>\n<head>\n\t<title>My Current Location</title>\n</head>\n\n<body>\n";

#Check to see if our file contains any data in it to read
if (-s $LocationStorageFileName > 0) {

	#There is some data, so attempt to open it for reading. 
	# An error will be printed if we could not read it.
	open(FILE,"<$LocationStorageFileName") or 
		&errorout("Could not read the location file. ".
		          "Check that you have given read permissions ".
			  "for the proper user.\n</body>\n</html>");

	#Request a SHARED lock on the file (since we are in read-only mode)
	flock FILE, 1;

	#Read in all the contents of the file delimiting by input record separators 
	# (which is CRLF on win32)
	@filecontents = <FILE>;

	#Close the file
	close(FILE);

	#Let everyone else waiting know we are done with the file.
	flock FILE, 8;

	#This prints out the information from the file. 
	# Feel free to play with the format. I have also linked the location to the 
	# Acme Mapper which loads images from TerraServer-USA. In this line I am 
	# explicitly assuming that the file will always be the format written to it 
	# above in this file.
	print "<table border=1><tr><td><h2>My Current Location is:<br><center>".
	      "<a href=\"http://mapper.acme.com/?lat=".
	      $filecontents[0]."&long=".$filecontents[1].
	      "&scale=10&width=4&height=3&dot=Yes\">$filecontents[0] ".
	      "Latitude<br>$filecontents[1] Longitude</a></center><br>I was at this ".
	      "location at this time:<br><center>$filecontents[2]".
	      "</center></h2></td></tr></table>";

} else {

	#If here, then there is no location file, or no content in that file, 
	# so print a generic message saying so.
	print "<h2>My Current Location is not available at this time</h2>";

}

#Close the standard HTML page
print "\n</body>\n</html>\n";


#This subroutine is only for errors while trying to read from or write to files. 
#It only prints a string sent to it and then exits.
sub errorout {
	my $errorstring = $_[0];
	print $errorstring;
	exit;
}
