	;
	; Ituner power sequencer
	;
	; pd 030611 initial code
	
	LIST	p=12F629 	;Target processor
	#include "P12F629.INC"	;Header file for target processor

	; GP0 = VSW12   input	1 = ignition turned on
	; GP1 = ON      output	1 = turn power on
	; GP2 = MBON#   o.d.	0 = power button pressed
	; GP3 = n.c.    input
	; GP4 = MBON2#  o.d.	0 = power button pressed (parallel out,
	;		makes connector flip-tolerant)
	; GP5 = n.c.    output	0 = terminate unused pin

	; Equates
	
PWR_OFF	equ	0x00	;power off state
PWR_ON	equ	0x02	;power on state

SW_ON	equ	0x09	;direction for switch on
SW_OFF	equ	0x1d	;direction for switch off

ONDELAY	equ	.20	;power on delay 4 seconds
ONDELAY2	equ	.5	;wait 1 second until push button
BUTTON	equ	.1	;push button for 400 ms
OFFDELAY	equ	.25	;shut down 5 s after ignition off
HARDOFF	equ	.225	;hard off after 45 seconds

	; Register variables

cntr	equ	0x20	;state counter
cntr2	equ	0x21	
waitl	equ	0x22	;delay counter low
waitm	equ	0x23	;delay counter mid
waith	equ	0x24	;delay counter high

	; reset entry

	org	0
	goto	reset
	
	; interrupt entry
	
	org	4
	retfie
	
reset
	banksel	OSCCAL	;get oscillator calibration
	call	0x3ff
	movwf	OSCCAL & 0x7f
	
	movlw	b'100000'	;enable weak pull-up on
	movwf	WPU & 0x7f	;GP5 (doesn't seem to work...)
	
	movlw	b'00001101'	;set option register
			;enable weak pull-ups
			;timer0 internal clock, positive edge
			;prescaler -> WDT, 1:32 -> ~0.5s
	movwf	OPTION_REG & 0x7f
	
reset2
	banksel	TRISIO
	movlw	SW_ON	;set pin directions
	movwf	TRISIO & 0x7f

	banksel	GPIO
	movlw	PWR_OFF	;set initial state
	movwf	GPIO
	
	movlw	7
	movwf	CMCON	;disable comparator
	
	clrwdt		;clear watchdog
	
	; initial state - wait until ignition is on for ONDELAY
	
initial	
	banksel	GPIO
	movlw	ONDELAY	;initial power on
	movwf	cntr
	
initwait	call	delay200	;wait
	btfss	GPIO,0	;ignition on ?
	  goto	initial	;no: clear counter
	decfsz	cntr,f	;initial time-out
	  goto	initwait
	
	; turn power on, wait ONDELAY2
	
turnon	movlw	PWR_ON
	movwf	GPIO
	movlw	ONDELAY2	;wait for ONDELAY2
	movwf	cntr
turnon1	call	delay200	;wait
	decfsz	cntr,f	
	  goto	turnon1
	btfss	GPIO,0	;ignition on ?
	  goto	reset2	;no: turn off again
	
	; push the system board power button

	banksel	TRISIO 	
	movlw	SW_ON	;push the button
	movwf	TRISIO & 0x7f
	banksel	GPIO
	movlw	BUTTON	;wait for BUTTON
	movwf	cntr
button1	call	delay200	;wait
	decfsz	cntr,f	
	  goto	button1
	banksel	TRISIO
	movlw	SW_OFF	;release the button
	movwf	TRISIO & 0x7f
	banksel	GPIO

	; power on state - shut down if ignition off for OFFDELAY

onstate	movlw	OFFDELAY
	movwf	cntr
onstate1	call	delay200
	btfsc	GPIO,0	;ignition on ?
	  goto	onstate	;yes: reset timer
	decfsz	cntr,f	
	  goto	onstate1
	
	; shut down - push the system board power button

	banksel	TRISIO	
	movlw	SW_ON	;push the button
	movwf	TRISIO & 0x7f
	banksel	GPIO
	movlw	BUTTON	;wait for BUTTON
	movwf	cntr
button2	call	delay200	;wait
	decfsz	cntr,f	
	  goto	button2
	banksel	TRISIO
	movlw	SW_OFF	;release the button
	movwf	TRISIO & 0x7f
	banksel	GPIO

	; turn-off state - turn off power after HARDOFF
	; go back to restart after ONDELAY

	movlw	HARDOFF
	movwf	cntr
	movlw	ONDELAY
	movwf	cntr2
shut1	call	delay200
	btfss	GPIO,0	;ignition on ?
	  goto	shut2	;:no
	decfsz	cntr2,f
	  goto	turnon	;go turn-on if ignition on for ONDELAY
	goto	shut3

shut2	movlw	ONDELAY	;ignition off - restart on timer
	movwf	cntr2
shut3	decfsz	cntr,f
	  goto	shut1	;keep waiting
	goto	reset2	;do hard power off

	; 200 ms delay routine

delay200	clrwdt		;tickle the watchdog
	movlw	.1	;note count is +1 (decfsz !)
	movwf	waith
	movlw	.0
	movwf	waitm
	clrf	waitl
del1	decfsz	waitl,f	;delay loop - 3 us per iteration
	  goto	del1
	decfsz	waitm,f
	  goto	del1
	clrwdt
	decfsz	waith,f
	  goto	del1
	retlw	0

	; oscillator calibration - comment out for production !!!

;	org	0x3ff
;	retlw	0x94

	; define configuration bits

	__config _MCLRE_OFF & _CP_OFF & _WDT_ON & _PWRTE_ON & _INTRC_OSC_NOCLKOUT

	; external reset off
	; code protection off (not read protected)
	;   note: if code protection is switched from on to off, OSCCAL
	;   value is erased too !!!
	; watchdog timer enabled
	; data memory protection off (not used)
	; internal RC oscillator
	; brown-out detect enable
	; power-up timer enable
	
	END
